#include <windows.h>
#include <stdio.h>
#include <math.h>
#include "LFSOpenVR.h"
#include "..\OpenVR\Include\OpenVR.h"

#if defined(_MSC_VER)
    #pragma warning(push, 0)
#endif
#include <d3d11.h>
#if defined(_MSC_VER)
    #pragma warning(pop)
#endif


// UTILS

inline int max_int(int i, int j)	{ return i < j ? j : i; }
inline int min_int(int i, int j)	{ return i > j ? j : i; }

void copy_str_zz(char *dst, ccs s, int max_len)
{
	int len = 0;
	ccs n = s;
	char *d = dst;
	while (*n && len < max_len-1)
	{
		*d = *n;
		d++;
		n++;
		len++;
	}

	while (len < max_len)
	{
		*d = 0;
		d++;
		len++;
	}
}

struct Vector
{
	float x;
	float y;
	float z;

	Vector(float ix, float iy, float iz) { x = ix; y = iy; z = iz; }
};


// STATIC

int						ovr_opened;

vr::IVRSystem			*ivr_system;
vr::IVRCompositor		*ivr_compositor;
vr::IVRControlPanel		*ivr_control_panel;

vr::VRTextureBounds_t	texture_bounds[2];

vr::TrackedDevicePose_t tracking_pose;

int						render_tex_width;	// the resulting width and height of the shared render texture
int						render_tex_height;


// D3D

ID3D11Device			*pDevice;
ID3D11DeviceContext		*pContext;
D3D_FEATURE_LEVEL		fl;

ID3D11Texture2D			*RenderTex;		// pointer to the shared render target
//ID3D11Texture2D			*pTexture2D;	// intermediate texture for submission - SRGB


// FUNCTIONS

LFSVR_API int LFSVR_GetVersion()
{
	return LFSVR_VERSION;
}

LFSVR_API int LFSVR_Open(OVR_HMDStart *start_info, int *error)
{
	*error = OR_FAIL_NONE;

	if (ovr_opened) return LFSVR_OPEN_FAIL;

	vr::EVRInitError init_error = vr::VRInitError_None;

	ivr_system = vr::VR_Init(&init_error, vr::VRApplication_Scene);

	if (ivr_system==NULL)
	{
		*error = OR_FAIL_INITIALIZE;
		return 0;
	}

	ivr_compositor = vr::VRCompositor();

	ivr_control_panel = vr::VRControlPanel();

	if (ivr_compositor==NULL || ivr_control_panel==NULL)
	{
		vr::VR_Shutdown();
		*error = OR_FAIL_HMD_CREATE;
		return 0;
	}

//	vr::TrackingUniverseOrigin g_origin = ivr_compositor->GetTrackingSpace();
	ivr_compositor->SetTrackingSpace(vr::TrackingUniverseSeated);

	// Initialize D3D11 and return an error code if it fails

	DWORD createDeviceFlags = 0;

//#ifdef _DEBUG
//	createDeviceFlags |= D3D11_CREATE_DEVICE_DEBUG;
//#endif

	// TODO : pAdapter should depend on the value of luid returned by ovr_Create

	HRESULT hr = D3D11CreateDevice(
		NULL,						// _In_opt_		IDXGIAdapter		*pAdapter,
		D3D_DRIVER_TYPE_HARDWARE,	//				D3D_DRIVER_TYPE		DriverType,
		NULL,						//				HMODULE				Software,
		createDeviceFlags,			//				UINT				Flags,
		NULL,						// _In_opt_		D3D_FEATURE_LEVEL	*pFeatureLevels,
		0,							//				UINT				FeatureLevels,
		D3D11_SDK_VERSION,			//				UINT				SDKVersion,
		&pDevice,					// _Out_opt_	ID3D11Device		**ppDevice,
		&fl,						// _Out_opt_	D3D_FEATURE_LEVEL	*pFeatureLevel,
		&pContext);					// _Out_opt_	ID3D11DeviceContext	**ppImmediateContext

	if (hr != S_OK)
	{
		vr::VR_Shutdown();
		*error = OR_FAIL_D3D11;
		return 0;
	}

	char buf[64];

	int len = ivr_system->GetStringTrackedDeviceProperty(0, vr::Prop_ModelNumber_String, buf, 64);
	ccs copy = len ? buf : "Device";
	copy_str_zz(start_info->ProductName, copy, 32);

	len = ivr_system->GetStringTrackedDeviceProperty(0, vr::Prop_ManufacturerName_String, buf, 64);
	copy = len ? buf : "OpenVR";
	copy_str_zz(start_info->Manufacturer, copy, 32);

	ovr_opened = 1;
	return 1; // ok
}

void CleanupTextures()
{
//	if (pTexture2D)
//	{
//		pTexture2D->Release();
//		pTexture2D = NULL;
//	}

	if (RenderTex)
	{
		RenderTex->Release();
		RenderTex = NULL;
	}
}

LFSVR_API void LFSVR_Close()
{
	if (ovr_opened==0) return; // nothing to do

	CleanupTextures();

	if (pContext) // release ID3D11DeviceContext
	{
		pContext->Release();
		pContext = NULL;
	}

	if (pDevice) // release ID3D11Device
	{
		pDevice->Release();
		pDevice = NULL;
	}

	vr::VR_Shutdown();
	ovr_opened = 0;
}

LFSVR_API int LFSVR_QueryHMD(OVR_HMDInfo *info, int want_tracking)
{
	if (ovr_opened==0) return 0; // can't do anything

	OVR_HMDInfo hmd_info;
	memset(&hmd_info, 0, sizeof(OVR_HMDInfo));

	hmd_info.Caps |= OR_TRACK_ORIENTATION;
	hmd_info.Caps |= OR_TRACK_POSITION;

	unsigned rec_w, rec_h;
	ivr_system->GetRecommendedRenderTargetSize(&rec_w, &rec_h);

	hmd_info.RTRecommendedW = rec_w * 2; // OpenVR recommended size is for one eye
	hmd_info.RTRecommendedH = rec_h;
	
	OVR_FovPort *dst = hmd_info.FovPort;
	for (int eye = 0; eye < 2; eye++)
	{
		vr::EVREye hmd_eye = eye ? vr::Eye_Right : vr::Eye_Left;

		float ltan, rtan, utan, dtan;

		ivr_system->GetProjectionRaw(hmd_eye, &ltan, &rtan, &utan, &dtan); // TODO - NOT SURE ABOUT THIS

		dst->LeftTan = fabsf(ltan);
		dst->RightTan = fabsf(rtan);
		dst->UpTan = fabsf(utan);
		dst->DownTan = fabsf(dtan);

		if (eye) // right
		{
			texture_bounds[eye].uMin = 0.5f;
			texture_bounds[eye].uMax = 1.0f;
		}
		else // left
		{
			texture_bounds[eye].uMin = 0.0f;
			texture_bounds[eye].uMax = 0.5f;
		}

		texture_bounds[eye].vMin = 0.0f;
		texture_bounds[eye].vMax = 1.0f;

		dst++;
	}

	hmd_info.IPD = ivr_control_panel->GetIPD();

	memcpy(info, &hmd_info, sizeof(OVR_HMDInfo));

	if (want_tracking==0) return LFSVR_OPEN_NO_SENSOR; // ok

	memset(&tracking_pose, 0, sizeof(vr::TrackedDevicePose_t));

	return LFSVR_OPEN_SENSOR;
}

const float OVR_SINGULARITYRADIUS = 0.0000001f;
const float OVR_PI = 3.14159265358979f;
const float OVR_PIOVER2 = 0.5f * OVR_PI;

LFSVR_API int LFSVR_GetTracking(float *yaw, float *pitch, float *roll, float *x, float *y, float *z)
{
	if (tracking_pose.bPoseIsValid==0) return 0;

	vr::HmdMatrix34_t *m = &tracking_pose.mDeviceToAbsoluteTracking;

	Vector r( m->m[0][0], -m->m[2][0],  m->m[1][0]);
	Vector f(-m->m[0][2],  m->m[2][2], -m->m[1][2]);
	Vector u( m->m[0][1], -m->m[2][1],  m->m[1][1]);

	if (f.z < -1.0f + OVR_SINGULARITYRADIUS) // down
	{
        *yaw	=  0.0f;
        *pitch	= -OVR_PIOVER2;
        *roll	=  atan2f(u.x, u.y);
	}
    else if (f.z > 1.0f - OVR_SINGULARITYRADIUS) // up
    {
        *yaw	=  0.0f;
        *pitch	=  OVR_PIOVER2;
        *roll	=  OVR_PI - atan2f(u.x, u.y);
	}
	else // usual case
	{
		*yaw	=  atan2f(-f.x, f.y);
		*pitch	=  asinf(f.z);
		*roll	= -atan2f(r.z, u.z);
	}

	*x =  m->m[0][3];
	*y = -m->m[2][3];
	*z =  m->m[1][3];

	return 2;
}

LFSVR_API void LFSVR_Reset()
{
	ivr_system->ResetSeatedZeroPose(); // NOT SURE
}

LFSVR_API ccs LFSVR_CreateTextures(int tex_width, int tex_height)
{
	render_tex_width = tex_width;
	render_tex_height = tex_height;

	D3D11_TEXTURE2D_DESC dsDesc;
	dsDesc.Width				= render_tex_width;
	dsDesc.Height				= render_tex_height;
	dsDesc.MipLevels			= 1;
	dsDesc.ArraySize			= 1;
	dsDesc.Format				= DXGI_FORMAT_B8G8R8A8_UNORM_SRGB;
	dsDesc.SampleDesc.Count		= 1;
	dsDesc.SampleDesc.Quality	= 0;
	dsDesc.Usage				= D3D11_USAGE_DEFAULT;
	dsDesc.BindFlags			= D3D11_BIND_SHADER_RESOURCE | D3D11_BIND_RENDER_TARGET;
	dsDesc.CPUAccessFlags		= 0;
	dsDesc.MiscFlags			= 0;

//	if (pDevice->CreateTexture2D(&dsDesc, NULL, &pTexture2D) != S_OK)
//	{
//		return "Could not create output texture";
//	}

	return NULL;
}

LFSVR_API ccs LFSVR_AcceptSharedTexture(HANDLE sharedHandle)
{
	if (RenderTex)
	{
		RenderTex->Release();
		RenderTex = NULL;
	}

	if (sharedHandle==NULL) return NULL;

	ID3D11Resource *tempResource11;
	HRESULT hr = pDevice->OpenSharedResource(sharedHandle, __uuidof(ID3D11Resource), (void**)(&tempResource11));

	if (hr != S_OK) return "AcceptSharedTexture : OpenSharedResource failed";
	
	hr = tempResource11->QueryInterface(__uuidof(ID3D11Texture2D), (void**)(&RenderTex));
	tempResource11->Release(); 

	if (hr != S_OK) return "AcceptSharedTexture : QueryInterface failed";

#ifdef _DEBUG

	D3D11_TEXTURE2D_DESC desc;
	RenderTex->GetDesc(&desc);

#endif

	return NULL;
}

LFSVR_API ccs LFSVR_SubmitFrame()
{
	if (RenderTex==NULL) return "SubmitFrame : RenderTex not supplied";

//	pContext->CopyResource(pTexture2D, RenderTex);

	vr::EVRCompositorError error;
	vr::Texture_t tex_info;

//	EColorSpace
//	ColorSpace_Auto = 0,	// Assumes 'gamma' for 8-bit per component formats, otherwise 'linear'.  This mirrors the DXGI formats which have _SRGB variants.
//	ColorSpace_Gamma = 1,	// Texture data can be displayed directly on the display without any conversion (a.k.a. display native format).
//	ColorSpace_Linear = 2,	// Same as gamma but has been converted to a linear representation using DXGI's sRGB conversion algorithm.

	tex_info.handle = RenderTex; // pTexture2D
	tex_info.eType = vr::API_DirectX;
	tex_info.eColorSpace = vr::ColorSpace_Gamma; // see EColorSpace above

	for (int eye = 0; eye < 2; eye++)
	{
		vr::EVREye hmd_eye = eye ? vr::Eye_Right : vr::Eye_Left;

		error = ivr_compositor->Submit(
			hmd_eye,
			&tex_info,
			texture_bounds + eye,
			vr::Submit_Default);

//		if (error && GetAsyncKeyState(VK_CONTROL)) return "Submit error";
	}

	pContext->Flush();

	error = ivr_compositor->WaitGetPoses(&tracking_pose, 1, NULL, 0);

//	if (error && GetAsyncKeyState(VK_CONTROL)) return "WaitGetPoses error";

	return NULL;
}