#ifndef _ORDIRECT_H_
#define _ORDIRECT_H_
////////////////////

#ifndef LFS_TYPES
#define LFS_TYPES
typedef unsigned char		byte;
typedef unsigned short		word;
typedef const char			*ccs;
#endif

#ifdef LFSVR_EXPORT
	#define LFSVR_API extern "C" __declspec(dllexport)
#endif

struct OVR_FovPort // same as ovrFovPort
{
    float UpTan;
    float DownTan;
    float LeftTan;
    float RightTan;
};

struct OVR_HMDStart
{
	char		ProductName		[32];
	char		Manufacturer	[32];

	unsigned	SpareU0;
	unsigned	SpareU1;
	unsigned	SpareU2;
	unsigned	SpareU3;
};

struct OVR_HMDInfo
{   
    unsigned	Caps;
    unsigned	Spare1;
    unsigned	Spare2;
	unsigned	Spare3;

	int			RTRecommendedW;
	int			RTRecommendedH;
	int			RTSpare2;
	int			RTSpare3;

	OVR_FovPort		FovPort[2]; // fields of view for each eye

	float		IPD;
	float		SpareF1;
	float		SpareF2;
	float		SpareF3;
};

// return values

const int LFSVR_VERSION = 1;

enum
{
	LFSVR_OPEN_FAIL,
	LFSVR_OPEN_START,
	LFSVR_OPEN_NO_SENSOR,
	LFSVR_OPEN_SENSOR,
	LFSVR_OPEN_NUM
};

enum
{
	OR_FAIL_NONE,
	OR_FAIL_INITIALIZE,
	OR_FAIL_HMD_CREATE,
	OR_FAIL_D3D11,
	OR_FAIL_NUM
};

#define OR_TRACK_ORIENTATION		0x00000100
#define OR_TRACK_POSITION			0x00000200

#ifdef LFSVR_EXPORT

LFSVR_API int		LFSVR_GetVersion			();

LFSVR_API int		LFSVR_Open					(OVR_HMDStart *start_info, int *error);
LFSVR_API void		LFSVR_Close					();
LFSVR_API int		LFSVR_QueryHMD				(OVR_HMDInfo *info, int want_tracking);

LFSVR_API int		LFSVR_GetTracking			(float *yaw, float *pitch, float *roll, float *x, float *y, float *z);
LFSVR_API void		LFSVR_Reset					(); // currently not used by LFS if position tracking is available

LFSVR_API ccs		LFSVR_CreateTextures		(int tex_width, int tex_height);
LFSVR_API ccs		LFSVR_AcceptSharedTexture	(HANDLE sharedHandle);
LFSVR_API ccs		LFSVR_SubmitFrame			();

#endif

//////
#endif